-- don't load if class is wrong
local _, myclass = UnitClass("player")
if myclass ~= "WARLOCK" then return end

-- EndarkenMe app
edm = {}

local LGT = LibStub:GetLibrary("LibGroupTalents-1.0", true)
if not LGT then
   print("Endarken Me is unable to load LibGroupTalents-1.0 so must exit.")
   return
end

-- WoW APIs
local GetTime = GetTime
local _ctime = GetTime()

edm.major = 0
edm.minor = 99
edm.build = "g"

-- Spell/Buff definitions
local i_darkintent_s			= 80398				-- Dark Intent (spell)
local i_darkintent_bt			= 85767       -- Dark Intent (buff to target)
local i_darkintent_bs			= 85768				-- Dark Intent (buff to self)
local i_darkintent_p			= 94310				-- Dark Intent (proc, stacks up to 3 times)

local n_darkintent	 			= GetSpellInfo(i_darkintent_s)
local b_di = GetSpellInfo(i_darkintent_bs)
local di_t = 0
local di_refresh_t = 10 * 60 * 1000        -- refresh when less than 10 minutes

-------------------------------------
-- Classes (non-localized)
-------------------------------------
local n_HUNTER				= "HUNTER"
local n_PALADIN       = "PALADIN"
local n_DRUID         = "DRUID"
local n_DEATHKNIGHT 	= "DEATHKNIGHT"
local n_ROGUE         = "ROGUE"
local n_SHAMAN        = "SHAMAN"
local n_WARRIOR       = "WARRIOR"
local n_WARLOCK       = "WARLOCK"
local n_PRIEST        = "PRIEST"
local n_MAGE          = "MAGE"

-------------------------------------
-- Talents by name (localized)
-------------------------------------
local n_Druid_talent_Balance      	= "Balance"
local n_Druid_talent_FeralCombat  	= "Feral Combat"
local n_Druid_talent_Restoration  	= "Restoration"
                                  	
local n_DeathKnight_talent_Blood  	= "Blood"
local n_DeathKnight_talent_Frost  	= "Frost"
local n_DeathKnight_talent_Unholy 	= "Unholy"
                                  	
local n_Paladin_talent_Holy       	= "Holy"
local n_Paladin_talent_Protection 	= "Protection"
local n_Paladin_talent_Retribution	= "Retribution"
                                  	
local n_Shaman_talent_Elemental   	= "Elemental"
local n_Shaman_talent_Enhancement 	= "Enhancement"
local n_Shaman_talent_Restoration 	= "Restoration"
                                  	
local n_Hunter_talent_BeastMastery	= "Beast Mastery"
local n_Hunter_talent_Marksmanship	= "Marksmanship"
local n_Hunter_talent_Survival      = "Survival"

local n_Mage_talent_Arcane          = "Arcane"
local n_Mage_talent_Fire            = "Fire"
local n_Mage_talent_Frost           = "Frost"

local n_Warrior_talent_Arms         = "Arms"
local n_Warrior_talent_Fury         = "Fury"
local n_Warrior_talent_Protection   = "Protection"

local n_Warlock_talent_Affliction   = "Affliction"
local n_Warlock_talent_Demonology   = "Demonology"
local n_Warlock_talent_Destruction  = "Destruction"

local n_Rogue_talent_Assasination   = "Assassination"
local n_Rogue_talent_Combat         = "Combat"
local n_Rogue_talent_Subtlety       = "Subtlety"

local n_Priest_talent_Discipline    = "Discipline"
local n_Priest_talent_Holy          = "Holy"
local n_Priest_talent_Shadow        = "Shadow"

--------------------------------------------------------------------------
-- Dark Intent Priority Lists
--------------------------------------------------------------------------
-- Priority based on Gherkin's Dark Intent thread:
-- http://www.mmo-champion.com/threads/813759-Mechanics-Dark-Intent-(4.2)

local RaidDpsGainDestruction = {
   { n_PRIEST,        n_Priest_talent_Shadow	    		},
   { n_DRUID,         n_Druid_talent_Balance         	},
   { n_MAGE,          n_Mage_talent_Fire             	},
   { n_DRUID,					n_Druid_talent_Restoration			},
   { n_SHAMAN,				n_Shaman_talent_Restoration			},
   { n_SHAMAN,        n_Shaman_talent_Elemental     	},
   { n_MAGE,          n_Mage_talent_Frost             },
   { n_MAGE,          n_Mage_talent_Arcane            },
   { n_DRUID,         n_Druid_talent_FeralCombat     	},
   { n_HUNTER,        n_Hunter_talent_Survival        },
   { n_PALADIN,       n_Paladin_talent_Retribution    },
   { n_ROGUE,         n_Rogue_talent_Combat           },
   { n_DEATHKNIGHT,   n_DeathKnight_talent_Unholy,	2	},    -- (2H)
   { n_ROGUE,         n_Rogue_talent_Assasination     },
   { n_DEATHKNIGHT,   n_DeathKnight_talent_Frost,		2	},    -- (2H)
   { n_HUNTER,        n_Hunter_talent_Marksmanship    },
   { n_HUNTER,        n_Hunter_talent_BeastMastery  	},
   { n_SHAMAN,        n_Shaman_talent_Enhancement    	},
   { n_DEATHKNIGHT,   n_DeathKnight_talent_Frost,		1	},    -- (1H)
   { n_DEATHKNIGHT,		n_DeathKnight_talent_Unholy,	1	},    -- (1H)
}

local RaidDpsGainAfflictionDrainLife = {
   { n_PRIEST,        n_Priest_talent_Shadow    			},
   { n_DRUID,         n_Druid_talent_Balance         	},
   { n_MAGE,          n_Mage_talent_Fire             	},
   { n_DRUID,					n_Druid_talent_Restoration			},
   { n_SHAMAN,				n_Shaman_talent_Restoration			},
   { n_SHAMAN,        n_Shaman_talent_Elemental     	},
   { n_MAGE,          n_Mage_talent_Frost             },
   { n_MAGE,          n_Mage_talent_Arcane            },
   { n_DRUID,         n_Druid_talent_FeralCombat     	},
   { n_HUNTER,        n_Hunter_talent_Survival        },
   { n_PALADIN,       n_Paladin_talent_Retribution    },
   { n_ROGUE,         n_Rogue_talent_Combat           },
   { n_HUNTER,        n_Hunter_talent_Marksmanship    },
   { n_DEATHKNIGHT,   n_DeathKnight_talent_Unholy,	2	},    -- (2H)
   { n_SHAMAN,        n_Shaman_talent_Enhancement    	},
   { n_DEATHKNIGHT,   n_DeathKnight_talent_Unholy,	1	},    -- (1H)
   { n_DEATHKNIGHT,   n_DeathKnight_talent_Frost,		2	},    -- (2H)
   { n_ROGUE,         n_Rogue_talent_Assasination     },
   { n_DEATHKNIGHT,   n_DeathKnight_talent_Frost,		1	},    -- (1H)
   { n_HUNTER,				n_Hunter_talent_BeastMastery  	},
}

local RaidDpsGainAfflictionShadowBolt = {
   { n_PRIEST,        n_Priest_talent_Shadow    			},
   { n_DRUID,         n_Druid_talent_Balance         	},
   { n_MAGE,          n_Mage_talent_Fire             	},
   { n_DRUID,					n_Druid_talent_Restoration			},
   { n_SHAMAN,				n_Shaman_talent_Restoration			},
   { n_SHAMAN,        n_Shaman_talent_Elemental     	},
   { n_MAGE,          n_Mage_talent_Frost             },
   { n_MAGE,          n_Mage_talent_Arcane            },
   { n_DRUID,         n_Druid_talent_FeralCombat     	},
   { n_HUNTER,        n_Hunter_talent_Survival        },
   { n_PALADIN,       n_Paladin_talent_Retribution    },
   { n_ROGUE,         n_Rogue_talent_Combat           },
   { n_SHAMAN,        n_Shaman_talent_Enhancement    	},
   { n_HUNTER,        n_Hunter_talent_Marksmanship    },
   { n_ROGUE,         n_Rogue_talent_Assasination     },
   { n_DEATHKNIGHT,   n_DeathKnight_talent_Frost,		1	},    -- (1H)
   { n_DEATHKNIGHT,   n_DeathKnight_talent_Frost,		2	},    -- (2H)
   { n_DEATHKNIGHT,   n_DeathKnight_talent_Unholy,	2	},    -- (2H)
   { n_HUNTER,        n_Hunter_talent_BeastMastery  	},
   { n_DEATHKNIGHT,   n_DeathKnight_talent_Unholy,	1	},    -- (1H)
}

local RaidDpsGainDemonology = {
   { n_PRIEST,        n_Priest_talent_Shadow    			},
   { n_DRUID,         n_Druid_talent_Balance         	},
   { n_MAGE,          n_Mage_talent_Fire             	},
   { n_DRUID,					n_Druid_talent_Restoration			},
   { n_SHAMAN,				n_Shaman_talent_Restoration			},
   { n_SHAMAN,        n_Shaman_talent_Elemental     	},
   { n_MAGE,          n_Mage_talent_Frost             },
   { n_MAGE,          n_Mage_talent_Arcane            },
   { n_DRUID,         n_Druid_talent_FeralCombat     	},
   { n_HUNTER,        n_Hunter_talent_Survival        },
   { n_PALADIN,       n_Paladin_talent_Retribution    },
   { n_ROGUE,         n_Rogue_talent_Combat           },
   { n_DEATHKNIGHT,   n_DeathKnight_talent_Unholy,	2	},    -- (2H)
   { n_ROGUE,         n_Rogue_talent_Assasination     },
   { n_DEATHKNIGHT,   n_DeathKnight_talent_Frost,		2	},    -- (2H)
   { n_HUNTER,        n_Hunter_talent_Marksmanship    },
   { n_HUNTER,        n_Hunter_talent_BeastMastery  	},
   { n_SHAMAN,        n_Shaman_talent_Enhancement    	},
   { n_DEATHKNIGHT,   n_DeathKnight_talent_Frost,		1	},    -- (1H)
   { n_DEATHKNIGHT,		n_DeathKnight_talent_Unholy,	1	},    -- (1H)
}

local PersonalUptime = {
   { n_DRUID,         n_Druid_talent_Balance					},
   { n_PRIEST,        n_Priest_talent_Shadow      		},
   { n_DRUID,					n_Druid_talent_Restoration			},
   { n_SHAMAN,				n_Shaman_talent_Restoration			},
   { n_HUNTER,        n_Hunter_talent_Survival    		},
   { n_DRUID,         n_Druid_talent_FeralCombat  		},
   { n_MAGE,          n_Mage_talent_Fire          		},
   { n_MAGE,          n_Mage_talent_Frost         		},
   { n_ROGUE,         n_Rogue_talent_Assasination 		},
   { n_SHAMAN,        n_Shaman_talent_Elemental   		},
   { n_DEATHKNIGHT,		n_DeathKnight_talent_Unholy,	1	},    -- (1H)
   { n_HUNTER,        n_Hunter_talent_BeastMastery		},
   { n_ROGUE,         n_Rogue_talent_Combat           },
   { n_MAGE,          n_Mage_talent_Arcane            },
   { n_DEATHKNIGHT,   n_DeathKnight_talent_Unholy, 	2 },    -- (2H)
   { n_HUNTER,        n_Hunter_talent_Marksmanship    },
   { n_SHAMAN,        n_Shaman_talent_Enhancement    	},
   { n_PALADIN,       n_Paladin_talent_Retribution    },
   { n_DEATHKNIGHT,   n_DeathKnight_talent_Frost, 	2 },    -- (2H)
   { n_DEATHKNIGHT,   n_DeathKnight_talent_Frost, 	1 },    -- (1H)
}

local AutoHide = false

-- Event Frame
local	ef = CreateFrame("Frame", nil, UIParent)      -- event frame

-- Update Frame and intervals
local uf = CreateFrame("Frame", nil, UIParent) 			-- update frame
local totalElapsed = 0
local oorElapsed = 0
local oorInterval = 1
local targetBuffElapsed = 0
local targetBuffInterval = 1

local groupdb = {}
local groupType = "party"
local prevMembers = 0
local numMembers = 1
local prilist = nil
local currTarget = nil
local prevTarget = nil
local oorQueue = {}				-- Out of Range party members
local inspectQueue = {}		-- In range members awaiting inspection
local inspectInProgress = false
local libGroupTalentUpdateComplete = false

-----------------------------------------------------------
-- vars for tracking Dark Intent stack uptime and stats
-----------------------------------------------------------
local totalCombat_t = 0
local beginCombat_t = 0

local procBarElapsed = 0
local buffBarElapsed = 0
local procExpiration = 0
local buffExpiration = 0
local buffRemindTime = 5
local buffTarget = "player"
local autoToggle = false

-- table to collect di proc stats
local didata = {}

-----------------------------------------------------------
-- UpdatePriList
-----------------------------------------------------------
function edm:UpdatePriList()
	--print("UpdatePriList()")
   
  --set up for party or raid

	if groupType == "raid" then   
	  --see what kind of warlock I am
	  local myspec = LGT:GetUnitTalentSpec("player")    
  	--print("I am a " .. myspec .. " Warlock")

  	-- pick priority list based on current spec
		if myspec == "Destruction" then
			prilist = RaidDpsGainDestruction
		elseif myspec == "Demonology" then
			prilist = RaidDpsGainDemonolgy
		elseif myspec == "Affliction" then
		  local ptsBane = LGT:UnitHasTalent("player", "Bane")
			if ptsBane == 3 then
				-- Bane 3/3 = Shadow Bolt
 				prilist = RaidDpsGainShadowBolt
			else
				-- Bane 0/3 = Drain Life
				prilist = RaidDpsGainAfflictionDrainLife
			end
		end
	else
   prilist = PersonalUptime
	end
end

-----------------------------------------------------------
-- SelectDarkIntentTarget
-----------------------------------------------------------
function edm:SelectDarkIntentTarget()
	local rank  = 41
	local realm = nil
	local WeaponType  = 0

	-- clear our selection
	currTarget = nil
   
	for k, v in pairs(groupdb) do
  	--print("checking priority for " .. v.name)
   	v.priority = 99
		for i = 1, #prilist do
			if v.inspected and v.class == prilist[i][1] and v.spec == prilist[i][2] then
	    	if v.class ~= n_DEATHKNIGHT then
   	    	--print(v.name, "is number", i, "on the", groupType, "priority list")
      	 	v.priority = i
        	if i < rank then
        		currTarget = v
        		rank = i
	        end
        else	-- it's a DK
					if v.spec ~= n_DeathKnight_talent_Blood then
			    	local weaponName, _, _, _, _, _, subclass = GetItemInfo(v.mainhandID)
			      print(v.name, "is wielding a " .. weaponName .. ", which is a", subclass)
			      if strmatch(subclass, "One") then
			      	WeaponType = 1
			      elseif strmatch(subclass, "Two") then
			      	WeaponType = 2
			      end
	          if WeaponType == prilist[i][3] then
		        	print(v.name, "is number", i, "on the", groupType, "priority list")
		 	       	v.priority = i
  	   			  if i < rank then
  	        		currTarget = v
			  	      rank = i
         			end
          	end
        	end
      	end
    	end
  	end -- i, #prilist do
	end -- k,v in pairs(groupdb)

--[[
	if currTarget then 
  	print("currTarget:", currTarget.name)
  else
  	print("no current target")
	end
	
	if prevTarget then
  	print("prevTarget:", prevTarget.name)
  else
  	print("no previous target")
  end
--]]		

	if currTarget then
		if currTarget ~= prevTarget then
	      prevTarget = currTarget
    		print(currTarget.name, "is the best Dark Intent target.")

			-- assign them to the unit attribute of the spell button and display name, spec, and class
			EdmFrame_Spell:SetAttribute("unit", currTarget.groupID)
    		EdmFrame_TargetName:SetText(currTarget.name..", "..currTarget.spec.." "..currTarget.lclass)

	    	-- build up name  for cross realm whisper if necessary
	    	local whisperChannel
  			if currTarget.realm then
    	  		whisperChannel = currTarget.name .. "-" .. currTarget.realm:gsub("%s", "")
    		else
	      	whisperChannel = currTarget.name
  	  		end
	    	
	    	--  whisper the recipient
  	  		local msg = "Hello "..currTarget.name..", Please stay near while I cast Dark Intent."
	      SendChatMessage(msg, "WHISPER", nil, whisperChannel)
    	else
    		print(currTarget.name, "is still the best Dark Intent target.")
    	end
	else
   	print("All of our choices are really bad.")
  	end
end -- SelectDarkIntentTarget()

-----------------------------------------------------------
-- RequestNextInspect
-----------------------------------------------------------
function edm:RequestNextInspect()
  --print("RequestNextInspect()")
  local m = table.remove(inspectQueue)
  if m then
    inspectInProgress = true
    NotifyInspect(m)
  else
    print("All inspections are complete, selecting Dark Intent target...")
      edm:SelectDarkIntentTarget()
    --end   
  end   
end

-----------------------------------------------------------
-- CheckOutOfRangeQueue
-----------------------------------------------------------
function edm:CheckOutOfRangeQueue()
	if #oorQueue == 0 then
		return
	end
	for i = 1, #oorQueue do
		local m = table.remove(oorQueue)
    if CheckInteractDistance(m, 1) then
	   	if not inspectInProgress then
  	 		inspectInProgress = true
     		NotifyInspect(m)
     	else
       	table.insert(inspectQueue, m)
      end
    else
    	table.insert(oorQueue, m)
    end
	end    
end

-----------------------------------------------------------
-- UpdateTargetBuff
-----------------------------------------------------------
function edm:UpdateTargetBuff()
	--print("UpdateTargetBuff()")
	if buffTarget then
		local _,_,_,_,_,_, expiration = UnitBuff(buffTarget, "Dark Intent", nil, "PLAYER")
		if expiration then
			buffExpiration = expiration
			local minutes = (expiration - _ctime - 1) / 60
			--print(UnitName(buffTarget), "has buff for", minutes, "more minutes")
		else
			buffExpiration = _ctime
			buffTarget = nil
		end
	end
end

-----------------------------------------------------------
-- UpdateAutoToggle
-----------------------------------------------------------
function edm:UpdateAutoToggle()
	if numMembers > 1 and buffTarget ~= currTarget.groupID then
		EdmFrame:Show()
	elseif ((buffExpiration - _ctime) / 60) < buffRemindTime then
		EdmFrame:Show()
	else
		EdmFrame:Hide()
	end
end

-----------------------------------------------------------
-- FrameOnUpdate
-----------------------------------------------------------
local function FrameOnUpdate(self, elapsed)

	-- update time on 1/100s		
  totalElapsed = totalElapsed + elapsed
  while (totalElapsed > 0.01) do
  	totalElapsed = totalElapsed - 0.01
   	_ctime = GetTime()
  end

	targetBuffElapsed = targetBuffElapsed + elapsed
  while (targetBuffElapsed > targetBuffInterval) do
  	targetBuffElapsed = targetBuffElapsed - targetBuffInterval
  	if buffTarget then
			edm:UpdateTargetBuff()
		end
		if autoToggle then
			edm:UpdateAutoToggle()
		end
  end

  oorElapsed = oorElapsed + elapsed
  while (oorElapsed > oorInterval) do
  	oorElapsed = oorElapsed - oorInterval
   	if #oorQueue > 0 then
   		edm:CheckOutOfRangeQueue()
   	end
  end
end

-----------------------------------------------------------
-- CleanGroupMemberData
-----------------------------------------------------------
function edm:CleanGroupMemberData()
	if groupdb then
		for k, v in pairs(groupdb) do
			if not (UnitInParty(v.name) or UnitInRaid(v.name)) then
				--print("removing", v.name, "from groupdb")
				if v == currTarget then
					print("Your Dark Intent target,", v.name, "has left the", groupType)
					currTarget = nil
				end
				groupdb[k] = nil
			end
		end
	end
end

-----------------------------------------------------------
-- UpdateGroupMemberData
-----------------------------------------------------------
function edm:UpdateGroupMemberData(getTalents)
	--print("UpdateGroupMemberData()")

	self:CleanGroupMemberData()

	if numMembers == 1 then
		return
	end

	for i = 1, numMembers do
		local m = groupType .. i
	   if UnitExists(m) and UnitInRaid(m) or UnitInParty(m) then
			-- create entry if it doesn't exist
			local guid = UnitGUID(m)
			if not groupdb[guid] then
				groupdb[guid] = {}
				--print("Obtaining information for ",  m)
				local name, realm = UnitName(m)
				local lclass, class = UnitClass(m)
				-- save player info in group db             
				groupdb[guid].groupID = m
				groupdb[guid].name = name
				groupdb[guid].realm = realm
				groupdb[guid].lclass = lclass
				groupdb[guid].class = class
				groupdb[guid].inspected = false
				groupdb[guid].priority = 41
				groupdb[guid].uptime = 0
				groupdb[guid].spec = "unknown"
			end

			if getTalents then
				--print("Obtaining talents for", m)
				local spec = LGT:GetUnitTalentSpec(m)
				groupdb[guid].spec = spec
				groupdb[guid].inspected = true
				
				if class == n_DEATHKNIGHT and spec ~= n_DeathKnight_talent_Blood then
					print(name .. " is a " .. spec .. " Death Knight, requesting inspect...")
					groupdb[guid].inspected = false
					if CheckInteractDistance(m, 1) then
						if not inspectInProgress then
							inspectInProgress = true
							NotifyInspect(m)
						else
							table.insert(inspectQueue, m)
						end
					else
						table.insert(oorQueue, m)
					end
				end
			end
   	end
  	end
end

-----------------------------------------------------------
-- HandleMemberChange
-----------------------------------------------------------
function edm:HandleMemberChange()
  --print("HandleMemberChange()")

  if UnitInRaid("player") then
  	numMembers = GetNumRaidMembers()
    groupType = "raid"
  elseif UnitInParty("party1") then
  	numMembers = GetNumPartyMembers() + 1
    groupType = "party"
	else
		numMembers = 1
		groupType = "none"
	end

  self:UpdatePriList()

	if prevMembers ~= numMembers then
		prevMembers = numMembers
  	if numMembers > 1 then
    	print("There are " .. numMembers .. " in the " .. groupType .. ".")
    else
    	print("I am all alone")
    	currTarget = nil
    	prevTarget = nil
    	groupdb = {}
    	EdmFrame_TargetName:SetText(UnitName("playerpet"))
			EdmFrame_Spell:SetAttribute("unit", "playerpet")
    end
	end
  
  if numMembers > 1 then
   if libGroupTalentUpdateComplete == true then
	  	self:UpdateGroupMemberData(true)
	  	if not inspectInProgess and #inspectQueue == 0 then
 				self:SelectDarkIntentTarget()
			end
		else
			self:UpdateGroupMemberData(false)
		end
	end
end

-----------------------------------------------------------
-- PARTY_MEMBERS_CHANGED
-----------------------------------------------------------
function edm:PARTY_MEMBERS_CHANGED()
	self:HandleMemberChange()
end

-----------------------------------------------------------
-- RAID_ROSTER_UPDATE
-----------------------------------------------------------
function edm:RAID_ROSTER_UPDATE()
	self:HandleMemberChange()
end

-----------------------------------------------------------
-- initDIdata
-----------------------------------------------------------
local function initDIdata(t)
	for i = 1, 3 do
		t[i] = {}
		t[i].beg_t = 0
		t[i].end_t = 0
		t[i].total_t = 0
	end
end	

-----------------------------------------------------------
-- UNIT_AURA
-----------------------------------------------------------
function edm:UNIT_AURA(unitID)
	--print("UNIT_AURA")
	_ctime = GetTime()
	local idx = 1

	repeat
		local name, rank, icon, count, debuffType, duration, expiration, unitCaster, isStealable, shouldConsolidate, spellID = UnitAura(unitID, idx, nil, nil) -- "PLAYER|HELPFUL") 
		if name == "Dark Intent" then			
		
			if spellID == i_darkintent_bt then
				local guid = UnitGUID(unitID)
				buffTarget = unitID
				buffExpiration = expiration
				if groupdb[guid] then
					local m = groupdb[guid]
	    		EdmFrame_TargetName:SetText(m.name..", "..m.spec.." "..m.lclass)
				else
	    		EdmFrame_TargetName:SetText(UnitName(unitID))
	    	end
			end

			if spellID == i_darkintent_bs then
					EdmFrame_BuffTime:SetValue((expiration - _ctime) / 60)
			end

			if spellID == i_darkintent_p then

				-- first time at this stack count			
				if didata[count].beg_t == 0 then
					didata[count].beg_t = expiration - duration
					didata[count].end_t = expiration
					procExpiration = expiration

					if count > 1 then
						EdmFrame_ProcTimeText:SetText(count.." Stacks")
					else
						EdmFrame_ProcTimeText:SetText("1 Stack")
					end
					
				-- already at this stack count, could be and extension but ignore repeats
				elseif didata[count].beg_t ~= 0 and expiration > didata[count].end_t then
					--print(count, "stack expiration extended by", format("%2.2f", expiration - didata[count].end_t), "seconds.")
					didata[count].end_t = expiration
					procExpiration = expiration
				end

				-- if we've bumped up a stack, stop counter on lower stack					
				if count > 1 and didata[count - 1].beg_t ~= 0  then
					didata[count - 1].end_t = didata[count].beg_t
				end
			end
		end
		idx = idx + 1
	until name == nil

		-- total out and clear out timers if we have passed expiration times
	for i = 1, 3 do
		if didata[i].end_t > 0 and didata[i].end_t <= _ctime then
			--print(i, "stack ends after", format("%2.2f", (didata[i].end_t - didata[i].beg_t)), "seconds.")
			if didata[i].end_t > 0 and didata[i].beg_t > 0 then
				didata[i].total_t = didata[i].total_t + (didata[i].end_t - didata[i].beg_t)
			end
			didata[i].beg_t = 0
			didata[i].end_t = 0
		end
	end
end

-----------------------------------------------------------
-- INSPECT_READY
-----------------------------------------------------------
function edm:INSPECT_READY(guid)
  --print("INSPECT_READY(", guid, ")")
  if groupdb[guid] and not groupdb[guid].inspected then
  	--print("Inspect ready for " .. groupdb[guid].name)
   	groupdb[guid].inspected = true
   	groupdb[guid].mainhandID = GetInventoryItemID(groupdb[guid].groupID, INVSLOT_MAINHAND)
   	local name, _, _, _, _, _, subclass = GetItemInfo(groupdb[guid].mainhandID)
   	ClearInspectPlayer()
   	inspectInProgress = false
  	self:RequestNextInspect()
  end
end

-----------------------------------------------------------
-- PLAYER_REGEN_ENABLED (ending combat)
-----------------------------------------------------------
function edm:PLAYER_REGEN_ENABLED()
	local elapsed = GetTime() - beginCombat_t
	--print("out of combat, elapsed time: ", format("%2.2f", elapsed), " seconds")
	totalCombat_t = totalCombat_t + elapsed
end

-----------------------------------------------------------
-- PLAYER_REGEN_DISABLED (beginning combat)
-----------------------------------------------------------
function edm:PLAYER_REGEN_DISABLED()
	beginCombat_t = GetTime()
	--print("in combat: ", beginCombat_t)
end

-----------------------------------------------------------
-- FrameOnEvent
-----------------------------------------------------------
local function FrameOnEvent(this, event, ...)
		--print("FrameOnEvent(" .. event .. ")")
    edm[event](edm, ...)
end



-----------------------------------------------------------
-- CreateFrame
-----------------------------------------------------------
function edm:InitFrames()

	-- set up event frame
	ef:RegisterEvent("UNIT_AURA")
	ef:RegisterEvent("INSPECT_READY")
	ef:RegisterEvent("PARTY_MEMBERS_CHANGED")
	ef:RegisterEvent("RAID__ROSTER_UPDATE")
	ef:RegisterEvent("PLAYER_REGEN_ENABLED")
	ef:RegisterEvent("PLAYER_REGEN_DISABLED")
	ef:SetScript("OnEvent", FrameOnEvent)
	
	-- set up update frame
	uf:SetScript("OnUpdate", FrameOnUpdate)
end


-----------------------------------------------------------
-- libGroupTalent events
-----------------------------------------------------------
-----------------------------------------------------------
-- LibGroupTalents_Add
-----------------------------------------------------------
function edm:LibGroupTalents_Add(self, guid, unit, name, realm)
	if unit ~= "player" then
		--print(name, "has arrived")
		local lclass, class = UnitClass(m)
		groupdb[guid] = {}
		groupdb.guid.groupID = unit
		groupdb.guid.name = name
		groupdb.guid.realm = realm
		groupdb.guid.class = class
		groupdb.guid.lclass = lclass
		groupdb.guid.inspected = false
		groupdb.guid.priority = 41
		groupdb.guid.uptime = 0
	  libGroupTalentUpdateComplete = false
	end
end

-----------------------------------------------------------
-- LibGroupTalents_Remove
-----------------------------------------------------------
function edm:LibGroupTalents_Remove(self, guid, unit, name, realm)
	local	_,_,_,_,_,name,realm = GetPlayerInfoByGUID(guid)	
	--print(name, "has departed")
	groupdb[guid] = nil
end

-----------------------------------------------------------
-- LibGroupTalents_Update
-----------------------------------------------------------
function edm:LibGroupTalents_Update(self, guid, unit, newSpec)
	-- update info in groupdb
	if unit ~= "player" and newSpec then
		--print(UnitName(unit), "has changed spec to", newSpec)
		groupdb[guid].spec = newSpec
		edm:SelectDarkIntentTarget()
	end
end

-----------------------------------------------------------
-- LibGroupTalents_UpdateComplete
-----------------------------------------------------------
function edm:LibGroupTalents_UpdateComplete(self, ...)
	local args = {...}
	for k, v in ipairs(args) do
		local _,_,_,_,_,name,realm = GetPlayerInfoByGUID(v)
		groupdb[guid].inspected = true
		--print(name, "has been updated")
	end
	libGroupTalentUpdateComplete = true
	self:UpdateGroupMemberData(true)
	self:SelectDarkIntentTarget()
end

-----------------------------------------------------------
-- slash command functions
-----------------------------------------------------------
-----------------------------------------------------------
-- HMS 
-----------------------------------------------------------
local function HMS(seconds)
	local h = 0
	local m = 0
	local s = 0
	
	if seconds and type(seconds) == "number" then
		h = floor(s / 3600)
		s = s - (h * 3600)
		m = floor(s / 60) 
		s = s - (m * 60)
	end
	return h, m, s
end

-----------------------------------------------------------
-- ShowTotals
-----------------------------------------------------------
function edm:ShowTotals()
	local totalDI_t = 0
	local percentDI = 0

	-- get total DI time		
	for i = 1, 3 do
		totalDI_t = totalDI_t + didata[i].total_t
	end
	
	if totalCombat_t > 0 then
		percentDI = totalDI_t / totalCombat_t * 100
	end		

	print("Total combat time:", format("%2.2f", totalCombat_t), "seconds")
	print("Total Dark Intent proc time:", format("%2.2f", totalDI_t), "seconds ("..format("%2.2f", percentDI).."% of combat time)")
	
	for i = 1, 3 do
		if didata[i].total_t > 0 then
			didata[i].percent = didata[i].total_t / totalDI_t * 100
		else
			didata[i].percent = 0
		end
		print("Total", i, "stack time:", format("%2.2f", didata[i].total_t), "seconds ("..format("%2.2f", didata[i].percent).."% of Dark Intent time)")
	end
end

-----------------------------------------------------------
-- ResetTotals
-----------------------------------------------------------
function edm:ResetTotals()
	print("Resetting Totals")
	totalCombat_t = 0
	initDIdata(didata)
	self:ShowTotals()
end

-----------------------------------------------------------
-- ShowPriority
-----------------------------------------------------------
function edm:ShowPriority()
	if prilist then
		print("Current Priority list is", groupType)
		print(" Rank    Class      Spec" )
		for i = 1, #prilist do
			print(i, prilist[i][1], prilist[i][2])
		end
	end
end

-----------------------------------------------------------
-- NumTableItems
-----------------------------------------------------------
local function NumTableItems(table)
	local count = 0
	if table then
		for k,v in pairs(table) do
			count = count + 1
		end
	end
	--print("NumTableItems returns", count)
	return count
end

-----------------------------------------------------------
-- DecimalToHex (color conversion)
-----------------------------------------------------------
local function DecimalToHex(r,g,b)
    return string.format("|cff%02x%02x%02x", r*255, g*255, b*255)
end

-----------------------------------------------------------
-- ShowGroup
-----------------------------------------------------------
function edm:ShowGroup()
	if NumTableItems(groupdb) > 0 then
		print("Current", groupType)
		for k, v in pairs(groupdb) do
			local color = RAID_CLASS_COLORS[v.class]
			print(DecimalToHex(color.r,color.g,color.b), v.groupID, v.name, v.spec, v.lclass, v.priority, v.inspected, "|r")
		end
	else
		print("There are no members in your group.")
	end
end

-----------------------------------------------------------
-- ShowTarget
-----------------------------------------------------------
function edm:ShowTarget()
	if currTarget then
		t = currTarget
		print(t.name, "a", t.spec, t.lclass, "is number", t.priority, "on the", groupType, "priority list.")
	else
		print("There is no Dark Intent target.")
	end
end

-----------------------------------------------------------
-- Toggle
-----------------------------------------------------------
function edm:Toggle()
	autoToggle = false
  if EdmFrame:IsShown() then
  	EdmFrame:Hide()
  else
  	EdmFrame:Show()
  end
end

-----------------------------------------------------------
-- Initialize
-----------------------------------------------------------
function edm:Initialize()

	initDIdata(didata)
	self:InitFrames()

	LGT.RegisterCallback(edm, "LibGroupTalents_UpdateComplete")
	LGT.RegisterCallback(edm, "LibGroupTalents_Update")
	LGT.RegisterCallback(edm, "LibGroupTalents_Add")
	LGT.RegisterCallback(edm, "LibGroupTalents_Remove")

	-- set up slash commands
	SLASH_EDM_OPTIONS1 = "/edm"
	SlashCmdList["EDM_OPTIONS"] = function(msg)
		msg = msg and string.lower(string.trim(msg))
		if msg == "totals" then
			edm:ShowTotals()
		elseif msg == "reset" then
			edm:ResetTotals()
		elseif msg == "group" then
			edm:ShowGroup()
		elseif msg == "party" then
			edm:ShowGroup()
		elseif msg == "raid" then
			edm:ShowGroup()
		elseif msg == "target" then
			edm:ShowTarget()
		elseif msg == "pri" or msg == "priority" then
			edm:ShowPriority()
		elseif msg == "toggle" then
			edm:Toggle()
		elseif msg == "auto" then
			autoToggle = not autoToggle
		elseif msg == "rescan" then
			edm:UpdateGroupMemberData()
			edm:SelectDarkIntentTarget()
		else
			print("I don't understand", msg)
		end
	end

	print("|c009482C9EndarkenMe version " .. edm.major .. "." .. edm.minor .. "." .. edm.build .. " loaded.|r")
end

-----------------------------------------------------------
-- Kick it all off here
-----------------------------------------------------------
edm:Initialize()



-----------------------------------------------------------
-- xml OnEvent, OnUpdate functions
-----------------------------------------------------------

-----------------------------------------------------------
-- edm:Button_OnClose
-----------------------------------------------------------
function edm:Button_OnClose(self, button)
	EdmFrame:Hide()
end

-----------------------------------------------------------
-- edm:Button_OnReport
-----------------------------------------------------------
function edm:Button_OnReport(self, button)
end


-----------------------------------------------------------
-- edm:Button_OnReportLoad
-----------------------------------------------------------
function edm:Button_OnReportLoad(self, button)
	self.icon:SetTexture("Interface\\Buttons\\UI-GuildButton-MOTD-Up")
end


-----------------------------------------------------------
-- EdmButton_OnLoad
-----------------------------------------------------------
function EdmSpellButton_OnLoad(button)

	local _,_,icon = GetSpellInfo(i_darkintent_s)

	button.icon:SetTexture(icon)	

if currTarget then
		button:SetAttribute("unit", currTarget.name)
	else
		button:SetAttribute("unit", "playerpet")
	end
	--self:RegisterForClicks("AnyDown")
	button:Show()
end


function EdmProcBar_OnLoad(bar)
--print("hello from proc bar")
end

-----------------------------------------------------------
-- Edm_ProcBarOnUpdate
-----------------------------------------------------------
function EdmProcBar_OnUpdate(bar, elapsed)
	local refresh = 0.05
	procBarElapsed = procBarElapsed + elapsed
  while (procBarElapsed > refresh) do
		procBarElapsed = procBarElapsed - refresh
		bar:SetValue(max(0, procExpiration - _ctime))
  end
end

-----------------------------------------------------------
-- EdmBuffBar_OnUpdate
-----------------------------------------------------------
function EdmBuffBar_OnUpdate(bar, elapsed)
	local refresh = 0.5
	buffBarElapsed = buffBarElapsed + elapsed
  while (buffBarElapsed > refresh) do
		buffBarElapsed = buffBarElapsed - refresh
		bar:SetValue(max(0, (buffExpiration - _ctime) / 60))
  end
end

-----------------------------------------------------------
-- EdmFrame
-----------------------------------------------------------
function EdmFrame_OnLoad(self)
	
 	EdmFrame_TargetName:SetText(UnitName("playerpet"))
	EdmFrame_Spell:SetAttribute("unit", "playerpet")

	EdmFrame_ProcTime:SetMinMaxValues(0, 7)
	EdmFrame_ProcTime:SetValue(5)
	EdmFrame_ProcTime:Show()
	
	EdmFrame_BuffTime:SetMinMaxValues(0, 30)
	EdmFrame_BuffTime:SetValue(15)
	EdmFrame_BuffTime:Show()

	self:Show()
	EdmFrameText:SetText("|c009482C9EndarkenMe version " .. edm.major .. "." .. edm.minor .. "." .. edm.build .."|r")
end
